@@ -4,10 +4,10 @@ import errno
 import json
 from concurrent.futures import ProcessPoolExecutor
 import io
-import re
 import time
 import pytest
 from itertools import chain
+import fsspec.core
 from s3fs.core import S3FileSystem
 from s3fs.utils import seek_delimiter, ignoring, SSEParams
 import moto
@@ -134,6 +134,21 @@ def test_simple(s3):
         assert out == data
 
 
+@pytest.mark.parametrize('default_cache_type', ['none', 'bytes', 'mmap'])
+def test_default_cache_type(s3, default_cache_type):
+    data = b'a' * (10 * 2 ** 20)
+    s3 = S3FileSystem(anon=False, default_cache_type=default_cache_type)
+
+    with s3.open(a, 'wb') as f:
+        f.write(data)
+
+    with s3.open(a, 'rb') as f:
+        assert isinstance(f.cache, fsspec.core.caches[default_cache_type])
+        out = f.read(len(data))
+        assert len(data) == len(out)
+        assert out == data
+
+
 def test_ssl_off():
     s3 = S3FileSystem(use_ssl=False)
     assert s3.s3.meta.endpoint_url.startswith('http://')
@@ -1211,6 +1226,7 @@ def test_change_defaults_only_subsequent():
         S3FileSystem.default_block_size = 5 * (1024 ** 2)
         S3FileSystem.cachable = True
 
+
 def test_passed_in_session_set_correctly(s3):
     session = boto3.session.Session()
     s3 = S3FileSystem(session=session)
@@ -1279,3 +1295,46 @@ def test_autocommit(s3):
     # Cannot commit a file that was discarded
     with pytest.raises(Exception):
         fo.commit()
+
+
+def test_touch(s3):
+    # create
+    fn = test_bucket_name + "/touched"
+    assert not s3.exists(fn)
+    s3.touch(fn)
+    assert s3.exists(fn)
+    assert s3.size(fn) == 0
+
+    # truncates
+    with s3.open(fn, 'wb') as f:
+        f.write(b'data')
+    assert s3.size(fn) == 4
+    s3.touch(fn, truncate=True)
+    assert s3.size(fn) == 0
+
+    # exists error
+    with s3.open(fn, 'wb') as f:
+        f.write(b'data')
+    assert s3.size(fn) == 4
+    with pytest.raises(ValueError):
+        s3.touch(fn, truncate=False)
+    assert s3.size(fn) == 4
+
+
+def test_seek_reads(s3):
+    fn = test_bucket_name + "/myfile"
+    with s3.open(fn, 'wb') as f:
+        f.write(b'a' * 175_627_146)
+    with s3.open(fn, 'rb', blocksize=100) as f:
+        f.seek(175561610)
+        d1 = f.read(65536)
+
+        f.seek(4)
+        size = 17562198
+        d2 = f.read(size)
+        assert len(d2) == size
+
+        f.seek(17562288)
+        size = 17562187
+        d3 = f.read(size)
+        assert len(d3) == size
